from typing import Union

# noinspection PyPackageRequirements,PyUnresolvedReferences
from gi.repository import GLib

import dbus
import dbus.service
import dbus.mainloop.glib

from platypush.backend import Backend
from platypush.context import get_bus
from platypush.message import Message
from platypush.message.event import Event
from platypush.message.request import Request
from platypush.utils import run


# noinspection PyPep8Naming
class DBusService(dbus.service.Object):
    @classmethod
    def _parse_msg(cls, msg: Union[dict, list]):
        import json
        return Message.build(json.loads(json.dumps(msg)))

    @dbus.service.method('org.platypush.MessageBusInterface', in_signature='a{sv}', out_signature='v')
    def Post(self, msg: dict):
        """
        This method accepts a message as a dictionary (either representing a valid request or an event) and either
        executes it (request) or forwards it to the application bus (event).

        :param msg: Request or event, as a dictionary.
        :return: The return value of the request, or 0 if the message is an event.
        """
        msg = self._parse_msg(msg)
        if isinstance(msg, Request):
            ret = run(msg.action, **msg.args)
            if ret is None:
                ret = ''  # DBus doesn't like None return types

            return ret
        elif isinstance(msg, Event):
            get_bus().post(msg)
            return 0


class DbusBackend(Backend):
    """
    This backend acts as a proxy that receives messages (requests or events) on the DBus and forwards them to the
    application bus.

    The name of the messaging interface exposed by Platypush is ``org.platypush.MessageBusInterface`` and it exposes
    ``Post`` method, which accepts a dictionary representing a valid Platypush message (either a request or an event)
    and either executes it or forwards it to the application bus.

    Requires:

        * **dbus-python** (``pip install dbus-python``)

    """

    def __init__(self, bus_name='org.platypush.Bus', service_path='/MessageService', *args, **kwargs):
        """
        :param bus_name: Name of the bus where the application will listen for incoming messages (default:
            ``org.platypush.Bus``).
        :param service_path: Path to the service exposed by the app (default: ``/MessageService``).
        """
        super().__init__(*args, **kwargs)
        self.bus_name = bus_name
        self.service_path = service_path

    def run(self):
        super().run()

        # noinspection PyUnresolvedReferences
        dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
        bus = dbus.SessionBus()
        name = dbus.service.BusName(self.bus_name, bus)
        srv = DBusService(bus, self.service_path)

        loop = GLib.MainLoop()
        # noinspection PyProtectedMember
        self.logger.info('Starting DBus main loop - bus name: {}, service: {}'.format(name._name, srv._object_path))
        loop.run()


# vim:sw=4:ts=4:et:



import enum
import json
from typing import Set, Dict, Optional
from xml.etree import ElementTree

import dbus

from platypush.plugins import Plugin, action


class BusType(enum.Enum):
    SYSTEM = 'system'
    SESSION = 'session'


class DbusPlugin(Plugin):
    """
    Plugin to interact with DBus.

    Requires:

        * **dbus-python** (``pip install dbus-python``)

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    @staticmethod
    def _get_bus_names(bus: dbus.Bus) -> Set[str]:
        return set([str(name) for name in bus.list_names() if not name.startswith(':')])

    @classmethod
    def path_names(cls, bus: dbus.Bus, service: str, object_path='/', paths=None, service_dict=None):
        if not paths:
            paths = {}

        paths[object_path] = {}
        obj = bus.get_object(service, object_path)
        interface = dbus.Interface(obj, 'org.freedesktop.DBus.Introspectable')
        xml_string = interface.Introspect()
        root = ElementTree.fromstring(xml_string)

        for child in root:
            if child.tag == 'node':
                if object_path == '/':
                    object_path = ''
                new_path = '/'.join((object_path, child.attrib['name']))
                cls.path_names(bus, service, new_path, paths)
            else:
                if not object_path:
                    object_path = '/'
                function_dict = {}
                for func in list(child):
                    if func.tag not in function_dict.keys():
                        function_dict[func.tag] = []
                    function_dict[func.tag].append(func.attrib['name'])

                if function_dict:
                    paths[object_path][child.attrib['name']] = function_dict

        if not service_dict:
            service_dict = {}
        if paths:
            service_dict[service] = paths

        return service_dict

    @action
    def query(self, service: Optional[str] = None, system_bus: bool = True, session_bus: bool = True) \
            -> Dict[str, dict]:
        """
        Query DBus for a specific service or for the full list of services.

        :param service: Service name (default: None, query all services).
        :param system_bus: Query the system bus (default: True).
        :param session_bus: Query the session bus (default: True).
        :return: A ``{service_name -> {properties}}`` mapping.
        """
        busses = {}
        response = {}

        if system_bus:
            busses['system'] = dbus.SystemBus()
        if session_bus:
            busses['session'] = dbus.SessionBus()

        for bus_name, bus in busses.items():
            services = {}
            service_names = self._get_bus_names(bus)

            if not service:
                for srv in service_names:
                    services[srv] = self.path_names(bus, srv)
            elif service in service_names:
                services[service] = self.path_names(bus, service)

            response[bus_name] = services

        return response

    @action
    def execute(self, service: str, path: str, method_name: str, args: Optional[list] = None,
                interface: Optional[str] = None, bus_type: str = BusType.SESSION.value):
        """
        Execute a method exposed on DBus.

        :param service: Service/bus name (e.g. ``org.platypush.Bus``).
        :param path: Object path (e.g. ``/MessageService``).
        :param method_name: Method name (e.g. ``Post``).
        :param args: Arguments to be passed to the method, depending on the method signature.
        :param interface: Interface name (e.g. ``org.platypush.MessageBusInterface``).
        :param bus_type: Bus type (supported: ``system`` and ``session`` - default: ``session``).
        :return: Return value of the executed method.
        """
        if not args:
            args = []

        kwargs = {}
        if interface:
            kwargs['dbus_interface'] = interface

        bus_type = BusType(bus_type)
        bus = dbus.SessionBus() if bus_type == BusType.SESSION else dbus.SystemBus()
        obj = bus.get_object(bus_name=service, object_path=path)
        ret = getattr(obj, method_name)(*args, **kwargs)
        return json.loads(json.dumps(ret))


# vim:sw=4:ts=4:et:



import json
import os
import pathlib

from platypush.plugins import Plugin, action


class FilePlugin(Plugin):
    """
    A plugin for general-purpose file methods
    """

    @classmethod
    def _get_path(cls, filename):
        return os.path.abspath(os.path.expanduser(filename))

    @classmethod
    def _to_string(cls, content):
        try:
            return json.dumps(content)
        except:
            return str(content)

    @action
    def read(self, file: str):
        """
        Read and return the content of a (text) file.

        :param file: Path of the file.
        """

        with open(self._get_path(file), 'r') as f:
            return f.read()

    @action
    def write(self, file: str, content: str):
        """
        Writes content to a specified (text) file. Previous content will be truncated.

        :param file: Path of the file.
        :param content: Content to write.
        """

        if not isinstance(content, str):
            content = self._to_string(content)

        with open(self._get_path(file), 'w') as f:
            f.write(content)

    @action
    def append(self, file: str, content):
        """
        Append content to a specified (text) file.

        :param file: Path of the file.
        :param content: Content to write.
        """

        if not isinstance(content, str):
            content = self._to_string(content)

        with open(self._get_path(file), 'a') as f:
            f.write(content)

    @action
    def getsize(self, file):
        """
        Get the size of the specified file in bytes.

        :param file: File path.
        """
        return os.path.getsize(self._get_path(file))

    @action
    def mkdir(self, directory: str, exist_ok=True, parents=True, mode=0o755):
        """
        Create a directory.

        :param directory: Directory name/path.
        :param exist_ok: If set and the directory already exist the method will
            not return an error (default: True).
        :param parents: If set and any of the parent directories in the path don't
            exist they will be created (analogous to mkdir -p) (default: True).
        :param mode: Access mode (default: 0755).
        """
        pathlib.Path(self._get_path(directory)).mkdir(parents=parents, exist_ok=exist_ok, mode=mode)

    @action
    def rmdir(self, directory: str):
        """
        Remove a directory. The directory must be empty.

        :param directory: Directory name/path.
        """
        pathlib.Path(self._get_path(directory)).rmdir()

    @action
    def touch(self, file: str, mode=0o644):
        """
        Create/touch a file.

        :param file: File name/path.
        :param mode: File permissions (default: 0644).
        """
        pathlib.Path(self._get_path(file)).touch(mode=mode)

    @action
    def chmod(self, file: str, mode):
        """
        Change the mode/permissions of a file.

        :param file: File name/path.
        :param mode: New file permissions.
        """
        pathlib.Path(self._get_path(file)).chmod(mode=mode)

    @action
    def home(self) -> str:
        """
        Returns the current user's home directory.
        """
        return str(pathlib.Path.home())

    @action
    def rename(self, file: str, name: str):
        """
        Rename/move a file.

        :param file: File to rename.
        :param name: New file name.
        """
        pathlib.Path(self._get_path(file)).rename(self._get_path(name))

    @action
    def link(self, file: str, target: str, symbolic=True):
        """
        Create a link to a file.

        :param file: File to symlink.
        :param target: Symlink path.
        :param symbolic: If True, then the target link will be a symbolic link. Otherwise,
            it will be a hard link (default: symbolic).
        """
        file = pathlib.Path(self._get_path(file))
        target = self._get_path(target)

        if symbolic:
            file.symlink_to(target)
        else:
            file.link_to(target)

    @action
    def unlink(self, file: str):
        """
        Remove a file or symbolic link.

        :param file: File/link to remove.
        """
        pathlib.Path(self._get_path(file)).unlink()


# vim:sw=4:ts=4:et:



"""
.. moduleauthor:: Fabio Manganiello <blacklight86@gmail.com>
.. license: MIT
"""

import logging
import re
import socket
import threading
import time

from threading import Thread
from typing import Optional

from platypush.bus import Bus
from platypush.config import Config
from platypush.context import get_backend
from platypush.message.event.zeroconf import ZeroconfServiceAddedEvent, ZeroconfServiceRemovedEvent
from platypush.utils import set_timeout, clear_timeout, \
    get_redis_queue_name_by_message, set_thread_name

from platypush import __version__
from platypush.event import EventGenerator
from platypush.message import Message
from platypush.message.event import Event, StopEvent
from platypush.message.request import Request
from platypush.message.response import Response


class Backend(Thread, EventGenerator):
    """
    Parent class for backends.

    A backend is basically a thread that checks for new events on some channel
    (e.g. a network socket, a queue, some new entries on an API endpoint or an
    RSS feed, a voice command through an assistant, a new measure from a sensor
    etc.) and propagates event messages to the main application bus whenever a
    new event happens. You can then build whichever type of custom logic you
    want on such events.
    """

    _default_response_timeout = 5

    # Loop function, can be implemented by derived classes
    loop = None

    def __init__(self, bus: Optional[Bus] = None, poll_seconds: Optional[float] = None, **kwargs):
        """
        :param bus: Reference to the bus object to be used in the backend
        :param poll_seconds: If the backend implements a ``loop`` method, this parameter expresses how often the
            loop should run in seconds.
        :param kwargs: Key-value configuration for the backend
        """

        self._thread_name = self.__class__.__name__
        EventGenerator.__init__(self)
        Thread.__init__(self, name=self._thread_name, daemon=True)

        # If no bus is specified, create an internal queue where
        # the received messages will be pushed
        self.bus = bus or Bus()
        self.poll_seconds = float(poll_seconds) if poll_seconds else None
        self.device_id = Config.get('device_id')
        self.thread_id = None
        self._should_stop = False
        self._stop_event = threading.Event()
        self._kwargs = kwargs
        self.logger = logging.getLogger(self.__class__.__name__)
        self.zeroconf = None
        self.zeroconf_info = None

        # Internal-only, we set the request context on a backend if that
        # backend is intended to react for a response to a specific request
        self._request_context = kwargs['_req_ctx'] if '_req_ctx' in kwargs \
            else None

        if 'logging' in kwargs:
            self.logger.setLevel(getattr(logging, kwargs.get('logging').upper()))

    def on_message(self, msg):
        """
        Callback when a message is received on the backend.
        It parses and posts the message on the main bus.
        It should be called by the derived classes whenever
        a new message should be processed.

        :param msg: Received message.  It can be either a key-value dictionary, a platypush.message.Message object,
            or a string/byte UTF-8 encoded string
        """

        msg = Message.build(msg)

        if not getattr(msg, 'target') or msg.target != self.device_id:
            return  # Not for me

        self.logger.debug('Message received on the {} backend: {}'.format(
            self.__class__.__name__, msg))

        if self._is_expected_response(msg):
            # Expected response, trigger the response handler
            clear_timeout()
            self._request_context['on_response'](msg)
            self.stop()
            return

        if isinstance(msg, StopEvent) and msg.targets_me():
            self.logger.info('Received STOP event on {}'.format(self.__class__.__name__))
            self._should_stop = True
        else:
            msg.backend = self   # Augment message to be able to process responses
            self.bus.post(msg)

    def _is_expected_response(self, msg):
        """ Internal only - returns true if we are expecting for a response
            and msg is that response """

        return self._request_context \
            and isinstance(msg, Response) \
            and msg.id == self._request_context['request'].id

    def _get_backend_config(self):
        config_name = 'backend.' + self.__class__.__name__.split('Backend')[0].lower()
        return Config.get(config_name)

    def _setup_response_handler(self, request, on_response, response_timeout):
        def _timeout_hndl():
            raise RuntimeError('Timed out while waiting for a response from {}'.
                               format(request.target))

        req_ctx = {
            'request': request,
            'on_response': on_response,
            'response_timeout': response_timeout,
        }

        resp_backend = self.__class__(bus=self.bus, _req_ctx=req_ctx,
                                      **self._get_backend_config(), **self._kwargs)

        # Set the response timeout
        if response_timeout:
            set_timeout(seconds=response_timeout, on_timeout=_timeout_hndl)

        resp_backend.start()

    def send_event(self, event, **kwargs):
        """
        Send an event message on the backend.

        :param event: Event to send. It can be a dict, a string/bytes UTF-8 JSON, or a platypush.message.event.Event
            object.
        """

        event = Event.build(event)
        assert isinstance(event, Event)

        event.origin = self.device_id
        if not hasattr(event, 'target'):
            event.target = self.device_id

        self.send_message(event, **kwargs)

    def send_request(self, request, on_response=None,
                     response_timeout=_default_response_timeout, **kwargs):
        """
        Send a request message on the backend.

        :param request: The request, either a dict, a string/bytes UTF-8 JSON, or a platypush.message.request.Request
            object.

        :param on_response: Optional callback that will be called when a response is received. If set, this method will
            synchronously wait for a response before exiting.
        :type on_response: function

        :param response_timeout: If on_response is set, the backend will raise an exception if the response isn't
            received within this number of seconds (default: None)
        :type response_timeout: float
        """

        request = Request.build(request)
        assert isinstance(request, Request)

        request.origin = self.device_id

        if on_response and response_timeout != 0:
            self._setup_response_handler(request, on_response, response_timeout)

        self.send_message(request, **kwargs)

    def send_response(self, response, request, **kwargs):
        """
        Send a response message on the backend.

        :param response: The response, either a dict, a string/bytes UTF-8 JSON, or a
            :class:`platypush.message.response.Response` object.
        :param request: Associated request, used to set the response parameters that will link them
        """

        assert isinstance(response, Response)
        assert isinstance(request, Request)

        self.send_message(response, **kwargs)

    def send_message(self, msg, queue_name=None, **kwargs):
        """
        Sends a platypush.message.Message to a node.
        To be implemented in the derived classes. By default, if the Redis
        backend is configured then it will try to deliver the message to
        other consumers through the configured Redis main queue.

        :param msg: The message to send
        :param queue_name: Send the message on a specific queue (default: the queue_name configured on the Redis
            backend)
        """

        try:
            redis = get_backend('redis')
            if not redis:
                raise KeyError()
        except KeyError:
            self.logger.warning("Backend {} does not implement send_message " +
                                "and the fallback Redis backend isn't configured")
            return

        redis.send_message(msg, queue_name=queue_name)

    def run(self):
        """ Starts the backend thread. To be implemented in the derived classes if the loop method isn't defined. """
        self.thread_id = threading.get_ident()
        set_thread_name(self._thread_name)
        if not callable(self.loop):
            return

        while not self.should_stop():
            try:
                with self:
                    has_error = False

                    while not self.should_stop() and not has_error:
                        try:
                            self.loop()
                        except Exception as e:
                            has_error = True
                            self.logger.error(str(e))
                            self.logger.exception(e)
                        finally:
                            if self.poll_seconds:
                                time.sleep(self.poll_seconds)
                            elif has_error:
                                time.sleep(5)
            except Exception as e:
                self.logger.error('{} initialization error: {}'.format(self.__class__.__name__, str(e)))
                self.logger.exception(e)
                time.sleep(self.poll_seconds or 5)

    def __enter__(self):
        """ Invoked when the backend is initialized, if the main logic is within a ``loop()`` function """
        self.logger.info('Initialized backend {}'.format(self.__class__.__name__))

    def __exit__(self, exc_type, exc_val, exc_tb):
        """ Invoked when the backend is terminated, if the main logic is within a ``loop()`` function """
        self.on_stop()
        self.logger.info('Terminated backend {}'.format(self.__class__.__name__))

    def on_stop(self):
        """ Callback invoked when the process stops """
        self.unregister_service()

    def stop(self):
        """ Stops the backend thread by sending a STOP event on its bus """
        def _async_stop():
            evt = StopEvent(target=self.device_id, origin=self.device_id,
                            thread_id=self.thread_id)

            self.send_message(evt)
            self._stop_event.set()
            self.on_stop()

        Thread(target=_async_stop).start()

    def should_stop(self):
        return self._should_stop

    def wait_stop(self, timeout=None) -> bool:
        return self._stop_event.wait(timeout)

    def _get_redis(self):
        import redis

        redis_backend = get_backend('redis')
        if not redis_backend:
            self.logger.warning('Redis backend not configured - some ' +
                                'web server features may not be working properly')
            redis_args = {}
        else:
            redis_args = redis_backend.redis_args

        redis = redis.Redis(**redis_args)
        return redis

    def get_message_response(self, msg):
        try:
            redis = self._get_redis()
            response = redis.blpop(get_redis_queue_name_by_message(msg), timeout=60)
            if response and len(response) > 1:
                response = Message.build(response[1])
            else:
                response = None

            return response
        except Exception as e:
            self.logger.error('Error while processing response to {}: {}'.format(msg, str(e)))

    @staticmethod
    def _get_ip() -> str:
        """
        Get the IP address of the machine.
        """
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        s.connect(('8.8.8.8', 80))
        addr = s.getsockname()[0]
        s.close()
        return addr

    def register_service(self, port: Optional[int] = None, name: Optional[str] = None, udp: bool = False):
        """
        Initialize the Zeroconf service configuration for this backend.
        """
        try:
            from zeroconf import ServiceInfo, Zeroconf
            from platypush.plugins.zeroconf import ZeroconfListener
        except ImportError:
            self.logger.warning('zeroconf package not available, service discovery will be disabled.')
            return

        self.zeroconf = Zeroconf()
        srv_desc = {
            'name': 'Platypush',
            'vendor': 'Platypush',
            'version': __version__,
        }

        name = name or re.sub(r'Backend$', '', self.__class__.__name__).lower()
        srv_type = '_platypush-{name}._{proto}.local.'.format(name=name, proto='udp' if udp else 'tcp')
        srv_name = '{host}.{type}'.format(host=self.device_id, type=srv_type)

        if port:
            srv_port = port
        else:
            srv_port = self.port if hasattr(self, 'port') else None

        self.zeroconf_info = ServiceInfo(srv_type, srv_name,
                                         addresses=[socket.inet_aton(self._get_ip())],
                                         port=srv_port,
                                         weight=0,
                                         priority=0,
                                         properties=srv_desc)

        self.zeroconf.register_service(self.zeroconf_info)
        self.bus.post(ZeroconfServiceAddedEvent(service_type=srv_type, service_name=srv_name,
                                                service_info=ZeroconfListener.parse_service_info(self.zeroconf_info)))

    def unregister_service(self):
        """
        Unregister the Zeroconf service configuration if available.
        """
        if self.zeroconf and self.zeroconf_info:
            self.zeroconf.unregister_service(self.zeroconf_info)
            if self.zeroconf:
                self.zeroconf.close()

            if self.zeroconf_info:
                self.bus.post(ZeroconfServiceRemovedEvent(service_type=self.zeroconf_info.type,
                                                          service_name=self.zeroconf_info.name))
            else:
                self.bus.post(ZeroconfServiceRemovedEvent(service_type=None, service_name=None))

            self.zeroconf_info = None
            self.zeroconf = None


# vim:sw=4:ts=4:et:



import datetime
import json
import os
import subprocess
import tempfile

from platypush.plugins import action
from platypush.plugins.http.request import Plugin


class HttpWebpagePlugin(Plugin):
    """
    Plugin to handle and parse/simplify web pages.
    It used to use the Mercury Reader web API, but now that the API is discontinued this plugin is basically a
    wrapper around the `mercury-parser <https://github.com/postlight/mercury-parser>`_ JavaScript library.

    Requires:

        * **requests** (``pip install requests``)
        * **weasyprint** (``pip install weasyprint``), optional, for HTML->PDF conversion
        * **node** and **npm** installed on your system (to use the mercury-parser interface)
        * The mercury-parser library installed (``npm install @postlight/mercury-parser``)
    """

    _mercury_script = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'mercury-parser.js')

    def _parse(self, proc):
        output = ''

        with subprocess.Popen(proc, stdout=subprocess.PIPE, stderr=None) as parser:
            output = parser.communicate()[0].decode()

        return output

    @action
    def simplify(self, url, type='html', html=None, outfile=None):
        """
        Parse the content of a web page removing any extra elements using Mercury

        :param url: URL to parse.
        :param type: Input type. Supported types: html, markdown, text (default: html).
        :param html: Set this parameter if you want to parse some HTML content already fetched. Note
            that URL is still required by Mercury to properly style the output, but it won't be used
            to actually fetch the content.

        :param outfile: If set then the output will be written to the specified file
            (supported formats: pdf, html, plain (default)). The plugin will guess
            the format from the extension
        :return: dict

        Example if outfile is not specified::

            {
                "url": <url>,
                "title": <page title>,
                "content": <page parsed content>

            }

        Example if outfile is specified::

            {
                "url": <url>,
                "title": <page title>,
                "outfile": <output file absolute path>

            }

        """

        self.logger.info('Parsing URL {}'.format(url))
        proc = ['node', self._mercury_script, url, type]
        f = None

        if html:
            f = tempfile.NamedTemporaryFile('w+', delete=False)
            f.write(html)
            f.flush()
            proc.append(f.name)

        try:
            response = self._parse(proc)
        finally:
            if f:
                os.unlink(f.name)

        try:
            response = json.loads(response.strip())
        except Exception as e:
            raise RuntimeError('Could not parse JSON: {}. Response: {}'.format(str(e), response))

        self.logger.debug('Got response from Mercury API: {}'.format(response))
        title = response.get('title', '{} on {}'.format(
            'Published' if response.get('date_published') else 'Generated',
            response.get('date_published', datetime.datetime.now().isoformat())))

        content = response.get('content', '')

        if not outfile:
            return {
                'url': url,
                'title': title,
                'content': content,
            }

        outfile = os.path.abspath(os.path.expanduser(outfile))
        content = '''<h1>{title}</h1><div class="_parsed-content">{content}</div>'''.\
            format(title=title, content=content)

        style = '''
            body {
                font-size: 22px;
                font-family: 'Merriweather', Georgia, 'Times New Roman', Times, serif;
            }
            '''

        if outfile.lower().endswith('.pdf'):
            import weasyprint
            from weasyprint.fonts import FontConfiguration

            font_config = FontConfiguration()
            css = [weasyprint.CSS('https://fonts.googleapis.com/css?family=Merriweather'),
                   weasyprint.CSS(string=style, font_config=font_config)]

            weasyprint.HTML(string=content).write_pdf(outfile, stylesheets=css)
        else:
            content = '''
                <html>
                    <head>
                        <title>{title}</title>
                        <style>{style}</style>
                    </head>
                    <body>{{content}}</body>
                </html>
            '''.format(title=title, style=style, content=content)

            with open(outfile, 'w', encoding='utf-8') as f:
                f.write(content)

        return {
            'url': url,
            'title': title,
            'outfile': outfile,
        }


# vim:sw=4:ts=4:et:
